/*
Version2.1 vom 3.10.2011
21.9.2011
Messung von Wasserstand, Temperatur sowie einigen Triacspannungen
und digitale Eingnge.(Drehschalter)
Realisiert mit ATmega88 - 4.096MHz (!!!!!!)

Daten werden via RS232 zum Hauptprozessor bertragen.


           (RESET)  PC6 1  28 PC5   (Spannung Pumpe)
              (RX)  PD0 2  27 PC4   Netzspannung hinter Trschalter
                TX  PD1 3  26 PC3   Netzspannung
     Nulldurchgang  PD2 4  25 PC2   Spannung Motor
                    PD3 5  24 PC1   Spannung Tr
                    PD4 6  23 PC0   Thermometer
                    VCC 7  22 GND
                    GND 8  21 AREF
                Q1  PB6 9  20 AVCC
                Q1  PB7 10 19 PB5   DI4 (SCK)
                    PD5 11 18 PB4   DI3 (MISO)
                    PD6 12 17 PB3   DI2 (MOSI)
                    PD7 13 16 PB2   DI1
     Druckmessdose  PB0 14 15 PB1   DI0

Ausgnge:
  PD1  TX

Eingnge:
  PB0    Druckmessdose  W    (ICP1)
  PB1    DI0
  PB2    DI1
  PB3    DI2
  PB4    DI3
  PB5    DI4
  
  PC0    Thermometer    T
  PC1    Spannung Tr   A
  PC2    Spannung Motor B
  PC3    Netzspannung1  C
  PC4    Netzspannung2  D (hinter Trschalter)
  PC5    Spannung Pumpe E
  PD2    Nulldurchgang
  PD3    DI1 
  PD4    DI2
  PD5    DI3
  PD6    DI4
  PD7    DI5
  

T0 - 8Bit:  Verzgert 5ms nach Nulldurchgang; Triggert ADC
T1 - 16Bit: zhlt mit prescaler=8. Mit Input-Capture Unit. Druckmessdosenpulse auf ICP1(PB0).
T2 - 8Bit:  Unbenutzt

Wasserstand: Die Druckdose liefert ca.44,5Hz bei Umgebungsdruck und 34Hz bei maximalem Druck.

34 * 65536 = 2,2Millionen
Also brauche ich einen 2 MHz Takt auf T1.
Statt dessen nehme ich einen Prescaler von 64, der bei 4,096MHz 64KHz macht und ermittle
die Frequenz ber 32 Schwingungen. Damit habe ich etwa jede Sekunde einen neuen Wert,
der bereits gefiltert ist. 

*/



#define F_CPU 4096000UL  /* CPU clock in Hertz */

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <stdint.h>

#define U8 uint8_t
#define U16 uint16_t
#define I8 int8_t
#define I16 int16_t
#define I32 int32_t
#define U32 uint32_t

#define BIT0 1
#define BIT1 2
#define BIT2 4
#define BIT3 8
#define BIT4 16
#define BIT5 32
#define BIT6 64
#define BIT7 128


//Datenbertragung zum Zentralprozessor
typedef struct {
  U8  FF;  //Start-
  U8  FE;  //      Zeichen
  U16 wasserstand;
  I16 temperatur;  // in 1/10 Grad Celsius
  U8   a;    
  U8   b;    
  U8   c;    
  U8   d;    
  U8   e;    
  U8   digitalInputs;
  U8  cr;  
  U8  lf;  
}TkleinhirnRecord;



//Thermometerkalibrierung
U16 thermometer[]={
    0x67,   //0   
    0xa0,   //5
    0xce,   //10
    //Ab hier: gemessen
    0x00fb,  //15   
    0x0128,  //20
    0x015b,  //25   
    0x018f,  //30
    0x01c3,  //35   
    0x01fb,  //40
    0x022b,  //45   
    0x025d,  //50
    0x0284,  //55   
    0x02A9,  //60
    0x02d5,  //65  
    0x02f4,  //70
    0x0309,  //75  
    //Ab hier: extrapoliert
    0x033f,  //80  
    0x0354,  //85  
    0x03A8,  //90  
    0x03c9   //95   
};    



//Der Einfachheit halber: viele globale Variablen


//Wir haben insgesamt 1024 Bytes SRAM - reichlich Platz, um alle Variablen statisch zu halten
typedef struct {
  U16 wasserstand; //Dauer von 32 Perioden. Hoher Wasserstand=niedriger Wert. 0:Dose kaputt
  U8  wasserstandGemessen;
  U8  keinePulse;
  U16 temperatur[2];  //ADC-Wert 0..1023, nicht normiert in Grad. 
  U16 auAdcVal[6][2];
  U8  neueAdcMessung;
  U8  adcValSlot;
  U8  tempSlot;


}Tglob;

//Alle globalen Var in einer Struct deklarieren
volatile  Tglob g;




//--------------------------------------------------------
ISR(TIMER1_CAPT_vect){ //Messdosenpuls-Interrupt
//--------------------------------------------------------
//Messe Capture Wert nach 32 Pulsen
static U8  pulse=1;
static U16 start;    

  if (! --pulse){
    g.wasserstand = ICR1 - start;      
    start = ICR1;
    pulse = 32;
    g.wasserstandGemessen = 1;  
  }
  g.keinePulse=0;

}
//--------------------------------------------------------
ISR(TIMER1_OVF_vect){ //"Kein Messdosenpuls" Interrupt
//--------------------------------------------------------
  g.keinePulse++;

  if (g.keinePulse > 1){ //Zwei berlufe ohne Interrupt: Druckdose kaputt!
    g.wasserstand =0;  
    g.keinePulse=1; 
    g.wasserstandGemessen = 1;  
  }
}


//--------------------------------------------------------
void USART_Init(void){  //RS232 initialisieren
//--------------------------------------------------------
 
	/* Set baud rate: round(f_osc / (8*baudrate)) - 1  wenn U2X=1*/
	/* Set baud rate: round(f_osc / (16*baudrate)) - 1  wenn U2X=0*/
	// 4096000 / 8 / 9600 -1 = 52
	UBRR0H = 0;
	UBRR0L = 105; //4800 Baud bei 4.096MHz, U2X0=1

  UCSR0A=_BV(TXC0)|_BV(U2X0); //Clear TXC; U2X0=1

	/* Set frame format:          2stop bit   8data,*/
	UCSR0C = _BV(USBS0)| _BV(UCSZ01) | _BV(UCSZ00);

	// Enable Transmitter 
	UCSR0B = _BV(TXEN0);
}


//--------------------------------------------------------
ISR(INT0_vect){ //Nulldurchgang (50Hz Interrupt)
//--------------------------------------------------------
  TCNT0 = 0;           //Reset Timer0, der startet in 5ms den ADC
}


//--------------------------------------------------------
ISR(TIMER0_COMPA_vect){
//--------------------------------------------------------
  //ADC wird automatisch gestartet (Klappte aber nicht, also von Hand)
  ADCSRA|= _BV(ADSC);
}

//--------------------------------------------------------
ISR(ADC_vect){ //ADC hat neuen Messwert
//--------------------------------------------------------
static U16 tempSum=0;
static U8 tempCnt=0;
static U8 channel=0;
U16 messwert;
U8 newSlot;
U8 newTempSlot;

  PORTD|=128;
  messwert = ADC;
  newSlot = g.adcValSlot ^ 1; //Schreibe in anderen Slot als gelesen wird
  g.auAdcVal[channel][newSlot] = messwert;
  if (0 == channel){
    tempSum += messwert;
    tempCnt++;
    if (tempCnt>=32){
      newTempSlot = g.tempSlot ^ 1;
      g.temperatur[newTempSlot] = tempSum/32;
      g.tempSlot = newTempSlot;
      tempSum=0;
      tempCnt=0;
    }
  }

  channel++;
  if (channel > 5){
    g.adcValSlot = newSlot;//Slot umschalten
    channel = 0;
  }
  ADMUX  = _BV(REFS0) | channel; //RefS1=0 RefS0=1 (Uref=AVcc) ADLAR=0

  PORTD&=~128;
  
}

//--------------------------------------------------------
I16 celsius(U16 messwert){
//--------------------------------------------------------
U8 i;
I16 delta;
  if (messwert>0x03c9) return 999;
  if (messwert<0x70) return 0;

  i=1; 
  while (messwert > thermometer[i]) i++;
  delta = thermometer[i] - thermometer[i-1];

  return (i-1)*50 + (messwert - thermometer[i-1]) *50 / delta;
  
}

//--------------------------------------------------------
int main (void)
//--------------------------------------------------------
{
static U8 i;
static U8 drehschalter;
static U8 drehschalterBisher;
static U8 *ram;

static U8 *pSendChar;
static TkleinhirnRecord kleinhirnRecord;
static U8 countdown;
static U8 txFinished;

  cli();

  //Watchdog
  asm("wdr");
  MCUSR &= ~(1<<WDRF);
  WDTCSR |= (1<<WDCE | 1<<WDE);
  WDTCSR=(1<<WDE | 1<<WDP2 | 1<<WDP1); //Watchdog auf 1s


  //AD-Wandler
  ADMUX=0x40; //RefS1=0 RefS0=1 (Uref=AVcc) ADLAR=0  **  Bit 0..2=Channel
  ADCSRA= _BV(ADEN) |  _BV(ADATE) | _BV(ADIF) |_BV(ADIE) | _BV(ADPS2) | _BV(ADPS1) | _BV(ADPS0);
  //ADEN (ADSC) ADATE ADIF ADIE ADPS2..0=1.1.1  (Division Factor=128 => 128KHz ADC-Takt)

  ADCSRB= _BV(ADTS1) | _BV(ADTS2) ; // Trigger on Timer0 Compare Match A
  DIDR0 = 31; //Digitale Eingnge auf ADC0..4 abschalten.


  //Set up the 8-bit timer 0
  //------------------------------------------
  //Timer wird in 50Hz Interrupt rckgesetzt. OCRA triggert ADC
  TCCR0A = 0;  //  Normal Modus
  TCCR0B = _BV(CS02) | _BV(CS00); //Prescaler 1:1024, gibt 4 KHz
  OCR0A  = 20; //5ms (bei 4Khz)
  TCNT0  = 0;
  TIMSK0 = _BV(OCIE0A);
    
  //Set up the 16-bit timer 1 (Druckdose)
  //---------------------------------------
  //Druckdosen-Signal an ICP1 pin (Using the Input Capture Unit)

  TCCR1A =  0; //Normaler Timer/Counter
  TCCR1B =  _BV(ICNC1)  //Input Capture Noise Canceler
          | _BV(ICES1)  //Trigger on rising edge
          | _BV(CS12)
          //| _BV(CS11)  // Clock Select 100 => Teiler=256 (640000Hz)
          //| _BV(CS10)
          ;

/*
      CS12 CS11 CS10 Description
      0    0    0    No clock source (Timer/Counter stopped).
      0    0    1    clk /1 (No prescaling)
      0    1    0    clk /8 (From prescaler)
      0    1    1    clk /64 (From prescaler)
      1    0    0    clk /256 (From prescaler)
      1    0    1    clk /1024 (From prescaler)
      1    1    0    External clock source on T1 pin. Clock on falling edge.
      1    1    1    External clock source on T1 pin. Clock on rising edge.
*/


  //Timer Interrupt Mask
  TIMSK1 = _BV(ICIE1) //Timer 1 Input Capture
          |_BV(TOIE1) //Overflow Interrupt Enable
          ;


  //------------------------------------------
  //Timer 2 wird nicht benutzt
  TCCR2A =0;  //  Normal Modus
  TCCR2B = 0; //  Timer  aus.

  DDRB = 0; //Alles Eingnge
  PORTB = BIT1|BIT2|BIT3|BIT4|BIT5; //Pull-Ups auf PB1..5

  DDRD = BIT1; //Alles Eingnge bis auf TX
  PORTD = BIT3|BIT4|BIT5|BIT6|BIT7; //Pull-Up auf PD3..7


  //EICRA=_BV(ISC00); //External interrupt INT0 (PD2) bei wechselnder Flanke  (Nulldurchgang)
  EICRA=_BV(ISC01);  //External interrupt INT0 (PD2) bei fallender Flanke  (Nulldurchgang) (S.72)
  EIMSK=_BV(INT0);

  USART_Init();


  //Strom sparen
  PRR = _BV(PRTWI) | _BV(PRSPI) | _BV(PRTIM2);

    //Globale Variablen initialisieren
//    memset(&g, 0, sizeof(Tglob));               Da rastet der Chip aus
  ram=(U8 *)&g;
  for (i=0; i< sizeof(Tglob); ram[i++]=0);   //Mach ich's halt von Hand

  kleinhirnRecord.FF=0xff;
  kleinhirnRecord.FE=0xfe;
  kleinhirnRecord.wasserstand =   1;
  kleinhirnRecord.temperatur  = 99;
  kleinhirnRecord.a = 42;
  kleinhirnRecord.b = 42;
  kleinhirnRecord.c = 42;
  kleinhirnRecord.d = 42;
  kleinhirnRecord.e = 42;
  kleinhirnRecord.digitalInputs = 0;
  kleinhirnRecord.cr=13;  
  kleinhirnRecord.lf=10;  
  pSendChar = (U8*) &kleinhirnRecord;
  txFinished=0;
  
  sei(); //Interrupts EIN. Jetzt geht's los!

  asm("wdr");  //Watchdog fttern


  TIFR1 |=1; //Reset Overflow-Flag on Timer1
  TCNT1 = 0; 
  countdown=30; //Zu Beginn 30 Telegramme im Dauerfeuer raus
//----------------------------------------------------------------------------------
  while(1){
//----------------------------------------------------------------------------------
    asm("wdr");  //Watchdog fttern
    drehschalter = (PINB >> 1) & 31;

    if (txFinished){ //bisheriges Telegramm ist raus 
      if (    (drehschalter!=drehschalterBisher) //Beim Start: wenn Drehschalter gendert, sofort senden damit Display das anzeigen kann.
           || (0!=g.wasserstandGemessen)           //Passiert ca. 1x pro Sekunde
           || (0!=countdown)
         )
      { 
        if (g.wasserstandGemessen){ //Neuer Wasserstand? (Sync wegen 16-Bit Zugriff ntig)
          kleinhirnRecord.wasserstand =   g.wasserstand;
          g.wasserstandGemessen=0;
        }
        drehschalterBisher=drehschalter;
        kleinhirnRecord.FF=0xff;
        kleinhirnRecord.FE=0xfe;
        kleinhirnRecord.temperatur  = celsius(g.temperatur[g.tempSlot]);
        kleinhirnRecord.a = (g.auAdcVal[1][g.adcValSlot]>>2)  & 255;
        kleinhirnRecord.b = (g.auAdcVal[2][g.adcValSlot]>>2)  & 255;
        kleinhirnRecord.c = (g.auAdcVal[3][g.adcValSlot]>>2)  & 255;
        kleinhirnRecord.d = (g.auAdcVal[4][g.adcValSlot]>>2)  & 255;
        kleinhirnRecord.e = (g.auAdcVal[5][g.adcValSlot]>>2)  & 255;
        kleinhirnRecord.digitalInputs = drehschalter;
        kleinhirnRecord.cr=13;  
        kleinhirnRecord.lf=10;  

        pSendChar = (U8*) &kleinhirnRecord; 
        txFinished=0;

        if (countdown) countdown--;
      }
    }
 

      //Daten via RS232 bertragen
    if (UCSR0A & (1<<UDRE0)){   //Puffer bereit, wenn UDRE0 gesetzt
      if (pSendChar <= &kleinhirnRecord.lf){ //Nicht am Record-Ende
        UDR0 = *pSendChar++;
      }else{
        _delay_ms(25);
        txFinished=1;
      }
    } 

   }  //while 1
   return (0);
}

